#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>

#include <signal.h>
#include <errno.h>

#ifdef __WIN32__
# include <winsock2.h>
# include <time.h>
#else
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>
# include <netdb.h>
#endif


#ifdef __WIN32__
typedef int socklen_t;
#else
static struct sigaction act;

void handle_alarm(int sig) {
  fputs("timeout...\n", stderr);
}
#endif


void usage(void) {
  puts("Usage: sendudp [-r | -j addr] [-s start-end] server [filename.prg]");
}


int sock;
socklen_t socklen;
struct sockaddr_in addr_send;


static unsigned char counter = 1;


int sendpacket(unsigned char *packet, int len) {
  int reply;
  int timeoutctr = 0;
  int acked = 0;
  struct sockaddr_in addr_reply;
  unsigned char recvbuffer[256];
#ifdef __WIN32__
  u_long nonblock = 1;
  u_long blocking = 0;
  time_t t;

  t = time(NULL);
#endif

  packet[1] = counter;

  do {
    if (sendto(sock, packet, len, 0, (struct sockaddr *)&addr_send, sizeof(struct sockaddr_in)) < 0) {
      perror("Couldn't send datagram");
      return 3;
    }

    do {
#ifdef __WIN32__
      ioctlsocket(sock, FIONBIO, &nonblock);
      reply = recvfrom(sock, recvbuffer, sizeof(recvbuffer), 0, (struct sockaddr *)&addr_reply, &socklen);
      ioctlsocket(sock, FIONBIO, &blocking);
#else
      alarm(1);
      reply = recvfrom(sock, recvbuffer, sizeof(recvbuffer), 0, (struct sockaddr *)&addr_reply, &socklen);
      alarm(0);
#endif
      if (reply <= 0) {
	if (errno == EINTR) {
	  if (++timeoutctr >= 3) {
	    fputs("No reply from server\n", stderr);
	    return 3;
	  }
#ifdef __WIN32__
	} else if (WSAGetLastError() == WSAEWOULDBLOCK) {
	  if (t < time(NULL)) {
	    t = time(NULL);
	    if (++timeoutctr >= 3) {
	      fputs("No reply from server\n", stderr);
	      return 3;
	    }
	    Sleep(1);
	  }
#endif
	} else {
#ifdef __WIN32__
	  /* for some weird reason I don't get WSAEWOULDBLOCK errors
	     fprintf(stderr, "Unknown WinSock error: %d, %d, %d\n", reply, errno, WSAGetLastError()); */
	  if (t < time(NULL)) {
	    t = time(NULL);
	    if (++timeoutctr >= 3) {
	      fputs("No reply from server\n", stderr);
	      return 3;
	    }
	    Sleep(1);
	  }
#else
	  perror("Unknown network error");
	  return 3;
#endif
	}
      } else {
	if (recvbuffer[0] == packet[0] && recvbuffer[1] == packet[1]) {
	  acked = 1;
	} else {
	  fprintf(stderr, "Acked wrong packet: expected %02x:%02x got %02x:%02x\n",
		  packet[0], packet[1], recvbuffer[0], recvbuffer[1]);
	  fputs("Resending...\n", stderr);
	  acked = 0;
	}
      }
    } while (reply < 0);

  } while (acked == 0);

  ++counter;

  return 0;
}


int senddata(unsigned short addr, unsigned char *buffer, int len) {
  unsigned char sendbuffer[256];

  sendbuffer[0] = 'L';
  sendbuffer[2] = addr & 0xff;
  sendbuffer[3] = addr >> 8;
  sendbuffer[4] = len;
  memcpy(&sendbuffer[5], buffer, len);

  return(sendpacket(sendbuffer, len + 5));
}


int main(int argc, char **argv) {
  struct hostent *hp;
  unsigned char filebuffer[256];
  unsigned char p[4];
  unsigned short addr;
  FILE *fh;
  int len;
  int opt;
  char *filename = NULL, *hostname = NULL;
  char action = 'L';
  int run = 0;
  int jump = -1;
  int res;

#ifdef __WIN32__
  WSADATA wsaData;

  if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0) {
    printf("WinSock startup failed\n");
    return 1;
  }
  atexit((void (*)(void))WSACleanup);
#else
  act.sa_handler = handle_alarm;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  if (sigaction(SIGALRM, &act, NULL) < 0 ) {
    perror("Couldn't enable SIGALRM handler");
    exit(1);
  }
#endif

  while ((opt = getopt(argc, argv, "rj:s:")) != -1) {
    switch (opt) {

    case 'r':
      run = 1;
      if (jump >= 0) {
	fputs("Can't enable both run and jump\n", stderr);
	return 1;
      }
      break;

    case 'j':
      jump = strtol(optarg, NULL, 0);
      if (run) {
	fputs("Can't enable both run and jump\n", stderr);
	return 1;
      }
      break;

    case 's':
      fputs("Save has not been implemented", stderr);
      return 1;
      action = 'S';
      break;

    case 'h':
    default:
      usage();
      return(1);
  
    }
  }
  argc -= optind;
  argv += optind;

  if (argc < 1 || argc > 2) {
    usage();
    return 1;
  }
  hostname = argv[0];
  if (argc == 2) {
    filename = argv[1];
  }

  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("Couldn't open socket");
    return 2;
  }

  addr_send.sin_family = AF_INET;
  if ((hp = gethostbyname(hostname)) == 0) {
    perror("Unknown host");
    return 2;
  }
  memcpy(&addr_send.sin_addr, hp->h_addr, hp->h_length);

  addr_send.sin_port = htons(3172);

  socklen = sizeof(struct sockaddr_in);

  if (filename) {

    if ((fh = fopen(filename, "rb")) == NULL) {
      perror("Couldn't open infile");
      return 1;
    }

    if (fread(filebuffer, 1, 2, fh) != 2) {
      perror("Couldn't read load address");
      fclose(fh);
      return 2;
    }
    addr = filebuffer[0] | (filebuffer[1]<<8);

    printf("Sending %s to $%04x\n", filename, (int) addr);

    while ((len = fread(filebuffer, 1, 128, fh)) > 0) {
      res = senddata(addr, filebuffer, len);
      if (res) {
	fclose(fh);
	return res;
      }
      addr += len;
    }

    fclose(fh);

  }

  if (run) {
    puts("Running");
    p[0] = 'R';
    return sendpacket(p, 2);
  } else if (jump >= 0) {
    printf("Jumping to $%04x\n", jump);
    p[0] = 'J';
    p[2] = jump & 0xff;
    p[3] = jump >> 8;
    return sendpacket(p, 4);
  }
  

  return 0;
}
