	.include "macro.i"
	.include "cs8900a.i"


	.export main


	.import cs_packet_page
	.import cs_packet_data
	.import cs_rxtx_data
	.import cs_tx_cmd
	.import cs_tx_len

	.import cfg_ip
	.import cfg_mac


ptr	= $ae


	.bss

stack:		.res 1	; 2
len		= $57	; 9
sendlen		= $59
sendptr		= $5a
pc_mac:		.res 6	; 4
pc_ip:		.res 4	; 6
ip_len		= $fd	; 11
ip_id:		.res 2	; 5
ip_cksum	= $fb	; 14
packet_len	= $ff	; 5

;udp_len	= packet_data + 4
udp_dest_port	= packet_data + 2
udp_src_port	= packet_data


	.segment "BUFFER"

	.align 256

packet_data:	.res $100


	.code

; read data from the buffer, aborting to the main routine if the buffer is empty
read_word:
	lda len + 1
	bmi abort
	bne @gotdata
	lda len
	beq abort
@gotdata:
	lda len
	sec
	sbc #2
	sta len
	bcs :+
	dec len + 1
:	lda cs_rxtx_data
	ldx cs_rxtx_data + 1
	rts


purge:
:	jsr read_word			; read data until buffer is empty and read_word aborts
	jmp :-


abort:
	ldx stack
	txs
	; fall through to main

; wait for packets
main:
	tsx
	stx stack

	lda #$24			; check rx status
	sta cs_packet_page
	lda #$01
	sta cs_packet_page + 1

@wait:
	lda cs_packet_data + 1
	and #$0d
	beq @wait

	lda cs_rxtx_data + 1		; ignore status
	lda cs_rxtx_data

	lda cs_rxtx_data + 1		; read packet length
	sta len + 1
	lda cs_rxtx_data
	sta len

	ldy #3				; skip dest mac (me or broadcast)
	jsr read_multiword

	jsr read_pc_mac

	jsr read_word
;	cmp #$08
;	bne purge
	cpx #0
	beq ip
	cpx #6
	bne purge


arp:
	ldy #4				; discard first 6 bytes
	jsr read_multiword		; and read arp command

	cpx #1				; make sure it's a request
	bne purge

	jsr read_pc_mac			; read senders mac address

	jsr read_pc_ip			; read senders ip address

	ldy #5				; discard target mac and my ip
	jsr read_multiword		; this means we reply to all ARP requests :)

;	ldy #4				; discard target mac address
;	jsr read_multiword		; and read first two target ip bytes
;
;	cmp cfg_ip			; check if they're asking for my ip
;	bne @arp_purge
;	cpx cfg_ip + 1
;	bne @arp_purge
;	jsr read_word
;	cmp cfg_ip + 2
;	bne @arp_purge
;	cpx cfg_ip + 3
;	bne @arp_purge

	; here we should purge data in a friendly manner...

	lda #6
	ldx #42
	jsr startreply			; start sending reply packet

	ldy #6				; send header
:	lda arp_reply_header,y
	ldx arp_reply_header + 1,y
	jsr send_word
	dey
	dey
	bpl :-

	jsr send_my_mac
	jsr send_my_ip
	jsr send_pc_mac
	jsr send_pc_ip

	jmp main


arp_reply_header:
	.dbyt $0002			; reply
	.byte $06			; hw addr len
	.byte $04			; proto addr len
	.dbyt $0800			; proto addr type (IPv4)
	.dbyt $0001			; hw addr type (Ethernet)


ip:
	ldy #2				; discard version, ihl, tos
	jsr read_multiword		; read len
	stx ip_len
	sta ip_len + 1
	lda ip_len
	ldx ip_len + 1
	sec				; subtract ip header length from packet length
	sbc #20
	bcs :+
	dex
:
	sta packet_len
	cpx #0				; restrict ip packet data to 256 bytes
	bne ip_purge

	jsr read_word			
	sta ip_id
	stx ip_id + 1

	ldy #2				; flags, frag, ttl
	jsr read_multiword		; read protocol
	txa
	pha

	jsr read_word			; read checksum
	sta ip_cksum
	stx ip_cksum + 1

	jsr read_pc_ip

	ldy #2				; discard dest ip
	jsr read_multiword

	pla
	cmp #17
	beq udp
;	cmp #1
;	bne ip_purge
;	jmp icmp
ip_purge:
	jmp purge


udp:
	lda packet_len			; round size up for odd sized payload
	lsr
	bcc :+
	inc packet_len
:
	ldy #0
@read:
	jsr read_word
	sta packet_data,y
	txa
	iny
	sta packet_data,y
	iny
	cpy packet_len
	bne @read
@flush:
	lda len + 1
	bmi @done
	bne @gotdata
	lda len
	beq @done
@gotdata:
	jsr read_word
	jmp @flush
@done:
	ldax udp_dest_port		; check for port 3172
	cmp #$0c
	bne @main
	cpx #$64
	bne @main

	lda packet_data + 8
	cmp #$4c
	beq udp_load
	cmp #$4a
	beq udp_jump
@main:
	jmp main


; Load addrlo addrhi len data...
udp_load:
	jsr udp_ack
	dec $d020
	sei
	lda #$34
	sta $01
	lda packet_data + 10
	sta ptr
	lda packet_data + 11
	sta ptr + 1
	ldy #0
:	lda packet_data + 13,y
	sta (ptr),y
	iny
	cpy packet_data + 12
	bne :-
	lda #$37
	sta $01
	cli
	inc $d020
	jmp main


udp_jump:
	jsr udp_ack
	jmp (packet_data + 10)


udp_ack:
	ldx #14 + 20 + 8 + 2		; eth + ip + udp + cmd + id
	lda #0
	jsr startreply

	lda #0				; reset checksum
	sta ip_cksum
	sta ip_cksum + 1
	sta ip_id
	sta ip_id + 1

	lda #$45			; add ipv4 and header length
	ldx #0
	jsr cksum_add

	lda #30				; add length of ip packet
	sta ip_len
	tax
	lda #0
	adc #0
	sta ip_len + 1
	jsr cksum_add

	; skip id and flags

	lda #$80			; add ttl and proto
	ldx #$11
	jsr cksum_add

	ldax cfg_ip
	jsr cksum_add
	ldax cfg_ip + 2
	jsr cksum_add

	ldax pc_ip
	jsr cksum_add
	ldax pc_ip + 2
	jsr cksum_add

	lda ip_cksum
	eor #$ff
	sta ip_cksum
	lda ip_cksum + 1
	eor #$ff
	sta ip_cksum + 1

	lda #$45			; send ver, ihl, tos
	ldx #$00
	jsr send_word

	lda ip_len + 1			; send len
	ldx ip_len
	jsr send_word

	ldax ip_id			; send id
	jsr send_word

	ldax #0				; send flags
	jsr send_word

	lda #$80			; send ttl and proto
	ldx #$11
	jsr send_word

	ldax ip_cksum			; send checksum
	jsr send_word

	jsr send_my_ip
	jsr send_pc_ip


	ldax udp_dest_port		; flip src/dest port
	jsr send_word
	ldax udp_src_port
	jsr send_word

	lda ip_len			; send udp packet length
	sec
	sbc #20
	tax
	lda ip_len + 1
	sbc #0
	jsr send_word

	ldax #0				; skip udp checksum
	jsr send_word

	ldax packet_data + 8
	jsr send_word

	rts


send_word:
	sta cs_rxtx_data
	stx cs_rxtx_data + 1
	rts


; type in A, length in X
startreply:
	pha

	lda #$c9			; ask for buffer space
	sta cs_tx_cmd
	lda #0
	sta cs_tx_cmd + 1

	stx cs_tx_len			; set length
	sta cs_tx_len + 1

	lda #<pp_bus_status		; select bus status register
	sta cs_packet_page
	lda #>pp_bus_status
	sta cs_packet_page + 1

:	lda cs_packet_data + 1		; wait for space
	ldx cs_packet_data		; add timeout?
	lsr
	bcc :-

	jsr send_pc_mac
	jsr send_my_mac

	lda #$08
	sta cs_rxtx_data
	pla
	sta cs_rxtx_data + 1

	rts


send_my_ip:
	ldax #cfg_ip
	bne send_multiple_4

send_pc_ip:
	ldax #pc_ip

send_multiple_4:
	ldy #4
	bne send_multiple

send_my_mac:
	ldax #cfg_mac
	bne send_multiple_6

send_pc_mac:
	ldax #pc_mac

send_multiple_6:
	ldy #6

send_multiple:
	stax sendptr
	sty sendlen
	ldy #0
:	lda (sendptr),y
	sta cs_rxtx_data
	iny
	lda (sendptr),y
	sta cs_rxtx_data + 1
	iny
	cpy sendlen
	bne :-
	rts


read_multiword:
:	jsr read_word
	dey
	bne :-
	rts


read_pc_mac:
	ldy #0
:	jsr read_word
	sta pc_mac,y
	txa
	sta pc_mac + 1,y
	iny
	iny
	cpy #6
	bne :-
	rts


read_pc_ip:
	ldy #0
:	jsr read_word
	sta pc_ip,y
	txa
	sta pc_ip + 1,y
	iny
	iny
	cpy #4
	bne :-
	rts


cksum_add:
	pha
	txa
	clc
	adc ip_cksum + 1
	sta ip_cksum + 1
	pla
	adc ip_cksum
	sta ip_cksum
	bcc :+
	inc ip_cksum + 1
:	rts
