	.include "macro.i"
	.include "cs8900a.i"


	.export main


	.import cs_packet_page
	.import cs_packet_data
	.import cs_rxtx_data
	.import cs_tx_cmd
	.import cs_tx_len

	.import cfg_ip
	.import cfg_mac


ptr	= $2d


	.bss

stack:		.res 1	; 2
len		= $57	; 9
sendlen		= $59
sendptr		= $5a
;pc_mac:		.res 6	; 4
;pc_ip:		.res 4	; 6
;ip_len		= $fd	; 11
;ip_id:		.res 2	; 5
;ip_cksum	= $fb	; 14
packet_len	= $ff	; 5

;udp_len:	.res 2	; 1
;udp_dest_port:	.res 2	; 2
;udp_src_port:	.res 2	; 2


	.segment "BUFFER"

	.align 256

eth_packet	= *
dest_mac:	.res 6
pc_mac:		.res 6
protocol:	.res 2
eth_data	= *
ip_ver:		.res 1
ip_tos:		.res 1
ip_len:		.res 2
ip_id:		.res 2
ip_flags:	.res 2
ip_ttl:		.res 1
ip_proto:	.res 1
ip_cksum:	.res 2
pc_ip		= *
ip_src_addr:	.res 4
ip_dest_addr:	.res 4
udp_src_port:	.res 2
udp_dest_port:	.res 2
udp_len:	.res 2
udp_cksum:	.res 2
udp_data	= *

arp_packet	= eth_data
arp_hw_at	= eth_data
arp_proto_at	= eth_data + 2
arp_hw_len	= eth_data + 4
arp_proto_len	= eth_data + 5
arp_op		= eth_data + 6
arp_src_mac	= eth_data + 8
arp_src_ip	= eth_data + 14
arp_tgt_mac	= eth_data + 18
arp_tgt_ip	= eth_data + 24


	.code


arp_reply_header:
	.dbyt $0002			; reply
	.dbyt $0604			; addr len
	.dbyt $0800			; proto addr type (IPv4)
	.dbyt $0001			; hw addr type (Ethernet)


arp:
	lda #6
	ldx #42
	jsr startreply			; start sending reply packet

	ldy #6				; send header
:	lda arp_reply_header,y
	ldx arp_reply_header + 1,y
	jsr send_word
	dey
	dey
	bpl :-

	jsr send_my_mac
	jsr send_my_ip
	jsr send_pc_mac
	jsr send_pc_ip

	;jmp main


; wait for packets
main:
	lda #$24			; check rx status
	sta cs_packet_page
	lda #$01
	sta cs_packet_page + 1

@wait:
	lda cs_packet_data + 1
	and #$0d
	beq @wait

	lda cs_rxtx_data + 1		; ignore status
	lda cs_rxtx_data

	lda cs_rxtx_data + 1		; read packet length
	sta len + 1
	lda cs_rxtx_data
	sta len
	lsr				; round up len
	bcc :+
	inc len
:
	ldx #0
@readpacket:
	lda cs_rxtx_data
	sta eth_packet,x
	inx
	lda cs_rxtx_data + 1
	sta eth_packet,x
	inx
	cpx len
	bne @readpacket

	ldy len + 1
	beq @notoversized
@purge:
	ldx #128
:	lda cs_rxtx_data
	lda cs_rxtx_data + 1
	dex
	bne :-
	dey
	bne @purge
	beq main
@notoversized:

	lda protocol + 1
	cmp #6
	beq arp
	cmp #0
	bne main


ip:
	lda ip_proto
	cmp #17
	bne main


udp:
	lda udp_dest_port
	cmp #$0c			; check for port 3172
	bne main
	lda udp_dest_port + 1
	cmp #$64
	bne main

	lda udp_data
	cmp #$4c
	beq udp_load
	cmp #$4a
	beq udp_jump
	jmp main


; Load addrlo addrhi len data...
udp_load:
	jsr udp_ack
	dec $d020
	sei
	lda #$34
	sta $01
	lda udp_data + 2
	sta ptr
	lda udp_data + 3
	sta ptr + 1
	ldy #0
:	lda udp_data + 5,y
	sta (ptr),y
	iny
	cpy udp_data + 4
	bne :-
	lda #$37
	sta $01
	cli
	inc $d020
	jmp main


udp_jump:
	jsr udp_ack
	jmp (udp_data + 2)


udp_ack:
	ldx #14 + 20 + 8 + 2		; eth + ip + udp + cmd + id
	lda #0
	jsr startreply

	lda #0				; reset checksum
	sta ip_cksum
	sta ip_cksum + 1
	sta ip_id
	sta ip_id + 1

	lda #$45			; add ipv4 and header length
	ldx #0
	jsr cksum_add

	ldx #30				; add length of ip packet
	stx ip_len + 1
	lda #0
	sta ip_len
	jsr cksum_add

	; skip id and flags

	lda #$80			; add ttl and proto
	ldx #$11
	jsr cksum_add

	ldax cfg_ip
	jsr cksum_add
	ldax cfg_ip + 2
	jsr cksum_add

	ldax pc_ip
	jsr cksum_add
	ldax pc_ip + 2
	jsr cksum_add

	lda ip_cksum
	eor #$ff
	sta ip_cksum
	lda ip_cksum + 1
	eor #$ff
	sta ip_cksum + 1

	lda #$45			; send ver, ihl, tos
	ldx #$00
	jsr send_word

	ldax ip_len			; send len
	jsr send_word

	ldax ip_id			; send id
	jsr send_word

	ldax #0				; send flags
	jsr send_word

	lda #$80			; send ttl and proto
	ldx #$11
	jsr send_word

	ldax ip_cksum			; send checksum
	jsr send_word

	jsr send_my_ip
	jsr send_pc_ip


	ldax udp_dest_port		; flip src/dest port
	jsr send_word
	ldax udp_src_port
	jsr send_word

	lda ip_len + 1			; send udp packet length
	sec
	sbc #20
	tax
	lda ip_len
	sbc #0
	jsr send_word

	ldax #0				; skip udp checksum
	jsr send_word

	ldax udp_data
	jmp send_word


send_word:
	sta cs_rxtx_data
	stx cs_rxtx_data + 1
	rts


; length in A/X, type in Y
startreply:
	pha

	lda #$c9			; ask for buffer space
	sta cs_tx_cmd
	lda #0
	sta cs_tx_cmd + 1

	stx cs_tx_len			; set length
	sta cs_tx_len + 1

	lda #<pp_bus_status		; select bus status register
	sta cs_packet_page
	lda #>pp_bus_status
	sta cs_packet_page + 1

:	lda cs_packet_data + 1		; wait for space
	ldx cs_packet_data		; add timeout?
	lsr
	bcc :-

	jsr send_pc_mac
	jsr send_my_mac

	lda #$08
	sta cs_rxtx_data
	pla
	sta cs_rxtx_data + 1

	rts


send_my_ip:
	ldax #cfg_ip
	bne send_multiple_4

send_pc_ip:
	ldax #pc_ip

send_multiple_4:
	ldy #4
	bne send_multiple

send_my_mac:
	ldax #cfg_mac
	bne send_multiple_6

send_pc_mac:
	ldax #pc_mac

send_multiple_6:
	ldy #6

send_multiple:
	stax sendptr
	sty sendlen
	ldy #0
:	lda (sendptr),y
	sta cs_rxtx_data
	iny
	lda (sendptr),y
	sta cs_rxtx_data + 1
	iny
	cpy sendlen
	bne :-
	rts


cksum_add:
	pha
	txa
	clc
	adc ip_cksum + 1
	sta ip_cksum + 1
	pla
	adc ip_cksum
	sta ip_cksum
	bcc :+
	inc ip_cksum + 1
:	rts
